@@ -123,15 +123,18 @@ List GetUDPost(List y,
 
          for (i = 0; i < Ng[g]; i ++)
          {
-            NumericVector Like_cw(C);
+            NumericVector like_cw(C);
 
             for (c = 0; c < C; c ++)
-               Like_cw(c) = gamma(w, c) * MeasP(i, c);
+            {
+               like_cw[c] = log(gamma(w, c)) + log(MeasP(i, c));
+            }
 
-            Uprob[w] += log(sum(Like_cw));
+            Uprob[w] += log(sum(exp(like_cw - max(like_cw)))) +
+               max(like_cw);
 
-            for (c = 0; c < C; c ++)
-               Dprob_w(i, c) = Like_cw(c) / sum(Like_cw);
+            Dprob_w.row(i) = exp(like_cw - max(like_cw)) /
+               sum(exp(like_cw - max(like_cw)));
          }
 
          Dprob[w] = Dprob_w;
@@ -248,15 +251,18 @@ List GetUDPostX(List y,
 
          for (i = 0; i < Ng[g]; i ++)
          {
-            NumericVector Like_cw(C);
+            NumericVector like_cw(C);
 
             for (c = 0; c < C; c ++)
-               Like_cw(c) = gamma_w(i, c) * MeasP(i, c);
+            {
+               like_cw[c] = log(gamma_w(i, c)) + log(MeasP(i, c));
+            }
 
-            Uprob[w] += log(sum(Like_cw));
+            Uprob[w] += log(sum(exp(like_cw - max(like_cw)))) +
+               max(like_cw);
 
-            for (c = 0; c < C; c ++)
-               Dprob_w(i, c) = Like_cw(c) / sum(Like_cw);
+            Dprob_w.row(i) = exp(like_cw - max(like_cw)) /
+               sum(exp(like_cw - max(like_cw)));
          }
 
          Dprob[w] = Dprob_w;
@@ -427,9 +433,10 @@ List GetScoreX(List y,
                int C,
                int M,
                IntegerVector R,
-               int P)
+               int P,
+               bool coeff_inv)
 {
-   int i, g, c, m, k, p, pos, ind;
+   int i, g, g2, c, m, k, p, pos, ind;
 
    List ret;
    List bscore(G);
@@ -447,10 +454,25 @@ List GetScoreX(List y,
       List rho_g = rho[g];
 
       for (i = 0; i < Ng[g]; i ++)
+      {
          for (c = 0; c < C - 1; c ++)
-            for (p = 0; p < P; p ++)
-               bscore_g(i, g * P * (C - 1) + c * P + p) =
-                  x_g(i, p) * (post_g(i, c) - gamma_g(i, c));
+         {
+            bscore_g(i, g * P * (C - 1) + c * P) =
+               post_g(i, c) - gamma_g(i, c);
+
+            for (p = 1; p < P; p ++)
+            {
+               if (coeff_inv)
+                  for (g2 = 0; g2 < G; g2 ++)
+                     bscore_g(i, g2 * P * (C - 1) + c * P + p) =
+                        x_g(i, p) * (post_g(i, c) - gamma_g(i, c));
+               else
+                  bscore_g(i, g * P * (C - 1) + c * P + p) =
+                     x_g(i, p) * (post_g(i, c) - gamma_g(i, c));
+            }
+         }
+      }
+
 
       for (m = 0; m < M; m ++)
       {
@@ -517,15 +539,18 @@ NumericMatrix GetUDScore(List y,
 
          for (i = 0; i < Ng[g]; i ++)
          {
-            NumericVector Like_cw(C);
+            NumericVector like_cw(C);
 
             for (c = 0; c < C; c ++)
-               Like_cw[c] = gamma(w, c) * MeasP(i, c);
+            {
+               like_cw[c] = log(gamma(w, c)) + log(MeasP(i, c));
+            }
 
-            Uprob[w] += log(sum(Like_cw));
+            Uprob[w] += log(sum(exp(like_cw - max(like_cw)))) +
+               max(like_cw);
 
-            for (c = 0; c < C; c ++)
-               Dprob_w(i, c) = Like_cw(c) / sum(Like_cw);
+            Dprob_w.row(i) = exp(like_cw - max(like_cw)) /
+               sum(exp(like_cw - max(like_cw)));
          }
 
          Dprob[w] = Dprob_w;
@@ -544,13 +569,13 @@ NumericMatrix GetUDScore(List y,
             for (c = 0; c < C; c ++)
                value1 += Uprob[w] * Dprob_w(i, c);
 
+            if (w < W - 1 && i == 0)
+               score(g, w) += value1 - delta[w];
+
             for (c = 0; c < C; c ++)
             {
                double value = Uprob[w] * Dprob_w(i, c);
 
-               if (w < W - 1)
-                  score(g, w) += value;
-
                if (c < C - 1)
                   score(g, W - 1 + w * (C - 1) + c) +=
                      value - gamma(w, c) * value1;
@@ -572,9 +597,6 @@ NumericMatrix GetUDScore(List y,
                   }
                }
             }
-
-            if (w < W - 1)
-               score(g, w) -= delta[w];
          }
       }
    }
@@ -597,10 +619,11 @@ NumericMatrix GetUDScoreX(List y,
                           int Q,
                           int C,
                           int M,
-                          IntegerVector R)
+                          IntegerVector R,
+                          bool coeff_inv)
 {
-   int i, g, w, c, p, q, m, k, pos, ind;
-   NumericMatrix score(G, W + (W * P + Q) * (C - 1) + C * (sum(R) - M));
+   int i, g, w, w2, c, p, q, m, k, pos, ind;
+   NumericMatrix score(G, W - 1 + (W * P + Q) * (C - 1) + C * (sum(R) - M));
 
    for (g = 0; g < G; g ++)
    {
@@ -620,15 +643,18 @@ NumericMatrix GetUDScoreX(List y,
 
          for (i = 0; i < Ng[g]; i ++)
          {
-            NumericVector Like_cw(C);
+            NumericVector like_cw(C);
 
             for (c = 0; c < C; c ++)
-               Like_cw[c] = gamma_w(i, c) * MeasP(i, c);
+            {
+               like_cw[c] = log(gamma_w(i, c)) + log(MeasP(i, c));
+            }
 
-            Uprob[w] += log(sum(Like_cw));
+            Uprob[w] += log(sum(exp(like_cw - max(like_cw)))) +
+               max(like_cw);
 
-            for (c = 0; c < C; c ++)
-               Dprob_w(i, c) = Like_cw(c) / sum(Like_cw);
+            Dprob_w.row(i) = exp(like_cw - max(like_cw)) /
+               sum(exp(like_cw - max(like_cw)));
          }
 
          Dprob[w] = Dprob_w;
@@ -657,10 +683,20 @@ NumericMatrix GetUDScoreX(List y,
 
                if (c < C - 1)
                {
-                  for (p = 0; p < P; p ++)
+                  score(g, W - 1 + w * (C - 1) * P + c * P) +=
+                     value - gamma_w(i, c) * value1;
+
+                  for (p = 1; p < P; p ++)
                   {
-                     score(g, W - 1 + w * (C - 1) * P + c * P + p) +=
-                        (value - gamma_w(i, c) * value1) * x_g(i, p);
+                     if (coeff_inv)
+                     {
+                        for (w2 = 0; w2 < W; w2 ++)
+                           score(g, W - 1 + w2 * (C - 1) * P + c * P + p) +=
+                              (value - gamma_w(i, c) * value1) * x_g(i, p);
+                     }
+                     else
+                        score(g, W - 1 + w * (C - 1) * P + c * P + p) +=
+                           (value - gamma_w(i, c) * value1) * x_g(i, p);
                   }
 
                   for (q = 0; q < Q; q ++)
@@ -780,12 +816,13 @@ double GetUDlik(List y,
       {
          for (i = 0; i < Ng[g]; i ++)
          {
-            NumericVector Like_cw(C);
+            NumericVector like_cw(C);
 
             for (c = 0; c < C; c ++)
-               Like_cw[c] = gamma(w, c) * MeasP(i, c);
+               like_cw[c] = log(gamma(w, c)) + log(MeasP(i, c));
 
-            clike[w] += log(sum(Like_cw)) - log(std::numeric_limits<double>::max());
+            clike[w] += log(sum(exp(like_cw - max(like_cw)))) +
+               max(like_cw) - log(std::numeric_limits<double>::max());
          }
       }
 
@@ -856,11 +893,16 @@ double GetUDlikX(List y,
          for (i = 0; i < Ng[g]; i ++)
          {
             NumericVector Like_cw(C);
+            NumericVector like_cw(C);
 
             for (c = 0; c < C; c ++)
+            {
                Like_cw[c] = gamma_w(i, c) * MeasP(i, c);
+               like_cw[c] = log(gamma_w(i, c)) + log(MeasP(i, c));
+            }
 
-            clike[w] += log(sum(Like_cw)) - log(std::numeric_limits<double>::max());
+            clike[w] += log(sum(exp(like_cw - max(like_cw)))) +
+               max(like_cw) - log(std::numeric_limits<double>::max());
          }
       }
 
@@ -914,6 +956,77 @@ List GetDeriv(NumericMatrix post,
    return ret;
 }
 
+// Gradient / Hessian for prevalence in MGLCA (coeff.inv)
+// [[Rcpp::export]]
+List GetDeriv2(List post,
+               List x,
+               List gamma,
+               IntegerVector Ng,
+               int G,
+               int C,
+               int P)
+{
+   int i, g, c1, c2, p, q;
+   double ind;
+   NumericVector grad((G + P - 1) * (C - 1));
+   NumericMatrix hess((G + P - 1) * (C - 1), (G + P - 1) * (C - 1));
+   List ret;
+
+   for (g = 0; g < G; g ++)
+   {
+      NumericMatrix gamma_g = gamma[g];
+      NumericMatrix post_g = post[g];
+      NumericMatrix x_g = x[g];
+
+      for (i = 0; i < Ng[g]; i ++)
+      {
+         for (c1 = 0; c1 < C - 1; c1 ++)
+         {
+            grad[g * (C - 1) + c1] += (post_g(i, c1) - gamma_g(i, c1));
+
+            for (c2 = 0; c2 < C - 1; c2 ++)
+            {
+               if (c1 == c2) ind = 1.0;
+               else ind = 0.0;
+               hess(g * (C - 1) + c1, g * (C - 1) + c2) +=
+                  - gamma_g(i, c1) * (ind - gamma_g(i, c2));
+
+               for (p = 1; p < P; p ++)
+               {
+                  hess(g * (C - 1) + c1, G * (C - 1) + c2 * (P - 1) + p - 1) +=
+                     - x_g(i, p) * gamma_g(i, c1) * (ind - gamma_g(i, c2));
+                  hess(G * (C - 1) + c2 * (P - 1) + p - 1, g * (C - 1) + c1) +=
+                     - x_g(i, p) * gamma_g(i, c1) * (ind - gamma_g(i, c2));
+               }
+            }
+
+            for (p = 1; p < P; p ++) // gradient
+            {
+               grad[G * (C - 1) + c1 * (P - 1) + p - 1] +=
+                  x_g(i, p) * (post_g(i, c1) - gamma_g(i, c1));
+               for (c2 = 0; c2 < C - 1; c2 ++)
+               {
+                  if (c1 == c2) ind = 1.0;
+                  else ind = 0.0;
+
+                  for (q = 1; q < P; q ++) // hessian
+                  {
+                     hess(G * (C - 1) + c1 * (P - 1) + p - 1,
+                          G * (C - 1) + c2 * (P - 1) + q - 1) +=
+                        x_g(i, p) * x_g(i, q) *
+                        - gamma_g(i, c1) * (ind - gamma_g(i, c2));
+                  }
+               }
+            }
+         }
+      }
+   }
+
+   ret["grad"] = grad;
+   ret["hess"] = hess;
+   return ret;
+}
+
 // Delta update
 // [[Rcpp::export]]
 NumericVector UpDelta(NumericMatrix PostW)
